/*
---------------------------------------------------------------------------
 Algo10-1.c   C program for implementing Algorithm 10.1
 Algorithm translated to  C  by: Dr. Norman Fahrer
 IBM and Macintosh verification by: Daniel Mathews
 
 NUMERICAL METHODS: C Programs, (c) John H. Mathews 1995
 To accompany the text:
 NUMERICAL METHODS for Mathematics, Science and Engineering, 2nd Ed, 1992
 Prentice Hall, Englewood Cliffs, New Jersey, 07632, U.S.A.
 Prentice Hall, Inc.; USA, Canada, Mexico ISBN 0-13-624990-6
 Prentice Hall, International Editions:   ISBN 0-13-625047-5
 This free software is compliments of the author.
 E-mail address:       in%"mathews@fullerton.edu"
 
 Algorithm 10.1 (Finite-Difference Solution for the Wave Equation).
 Section   10.1, Hyperbolic Equations, Page 507
---------------------------------------------------------------------------
*/
/*
---------------------------------------------------------------------------
 
 Algorithm 10.1 (Finite-Difference Solution for the Wave Equation)

                                                 2
 To approximate the solution of u_(tt)(x,t)  =  c u_(xx)(x,t)

 over  R = {(x,t): 0 <= x <= a, 0 <= t <= b}  with 

 u(0,t) = 0,       u(a,t) = 0      for  0 <= t <= b and 

 u(x,0) = f(x),  u_t(x,0) = g(x)   for  0 <= x <= a.

 User has to supply function  Fi(gridpoint)  and 
 Gi(gridpoint):  boundary functions at the grid
 points. An examples is implemented in this program.

---------------------------------------------------------------------------
*/

#include<stdio.h>
#include<stdlib.h>
#include<math.h>

#define Max 50

/* Global Variables */

double H;                /* Step size  */

/* Prototypes */

double Fi(int i);
double Gi(int i);

/* Grid function for amplitude */

double Fi(int i)
{
    extern double H;
    double arg;

    arg = H * (i - 1);

    if( (arg >= 0) && (arg <= 1.0) )  return ( sin ( 3.1415926 * arg) );
    else
    {
        printf(" Fi() :Argument not in range ! Exiting !\n");
        printf(" arg  : %lf\n", arg);
        exit(0);
    }
}

/* Grid function for velocity */

double Gi(int i)
{
    extern double H;
    double arg;

    arg = H * (i - 1);

    if( (arg >= 0) && (arg <= 1) ) return ( 0.0 );
    else
    {
        printf(" Gi() :Argument not in range ! Exiting !\n");
        printf(" arg  : %lf\n", arg);
        exit(0);
    }
}

/* -------------------------------------------------------- */

/*  Main program for algorithm 10-1  */

void main(void)

{
    int i, j;                /* Loop Counters                         */
    double A, B;             /* INPUT :Width and height of Rectangle  */
    double C;                /* INPUT : Wave equation const.          */
    int N, M;                /* Dimensions of the grid                */
    double K, R, R2, R22, S1, S2;
    double U[Max][Max];      /* Grid-Amplitudes                       */

    printf("-- Finite-Difference Solution to the Wave Equation ----\n");
    printf("--------- Example Problem No 4 on page 508 ------------\n");
    printf("-------------------------------------------------------\n");

    printf("Please enter A (interval boundary of x)\n");
    printf("For present example type : 1.0\n");
    scanf("%lf",&A);
    printf("Please enter B (interval boundary of t)\n");
    printf("For present example type : 0.5\n");
    scanf("%lf",&B);
    printf("You entered A = %lf and B = %lf \n", A, B);
    printf("-------------------------------------------------------\n");
    printf("\n");
    printf("Please enter dimension of grid in x-direction :\n");
    printf("For present example type : 6\n");
    scanf("%d",&N);
    printf("Please enter dimension of grid in y-direction :\n");
    printf("For present example type : 6\n");
    scanf("%d",&M);
    printf("You entered N = %d and M = %d\n", N, M);
    printf("-------------------------------------------------------\n");
    printf("\n");
    printf("Please enter the wave equation constant C :\n");
    printf("For present example type : 2\n");
    scanf("%lf",&C);
    printf("You entered C = %lf\n", C);
    printf("------------------------------------------------\n");
    printf("\n");

    /* Compute step sizes */

    H   = A / ( N - 1 );
    K   = B / ( M - 1 );
    R   = C * K / H;
    R2  = R * R;
    R22 = R * R / 2.0;
    S1  = 1.0 - R * R;
    S2  = 2.0 - 2.0 * R * R;

    /* Boundary conditions */

    for ( j = 1; j <= M; j++ )
    {
        U[1][j] = 0;
        U[N][j] = 0;
    }

    /* First and second rows */

    for ( i = 2; i <= N - 1 ; i++ ) 
    {
        U[i][1] = Fi(i);
        U[i][2] = S1 * Fi(i) + K * Gi(i) + R22 * ( Fi(i+1) + Fi(i-1) );
    }

    /* Generate new waves */

    for ( j = 3; j <= M; j++ )
    {
        for ( i = 2; i <= N - 1; i++ )
        {
            U[i][j] = S2 * U[i][j-1] + R2 * ( U[i-1][j-1] + U[i+1][j-1] )
                         - U[i][j-2];
        }
    }
     
    /* Output the solution */

    for ( j = 1; j <= M; j++ )
    {
        printf("%8.6lf ", K * (j - 1));

        for ( i = 1; i <= N; i++ ) printf(" %8.6lf ", U[i][j]);

        printf("\n");
    }

}   /* End of main program */

