/*
---------------------------------------------------------------------------
 Algo10-4.c   C program for implementing Algorithm 10.4
 Algorithm translated to  C  by: Dr. Norman Fahrer
 IBM and Macintosh verification by: Daniel Mathews
 
 NUMERICAL METHODS: C Programs, (c) John H. Mathews 1995
 To accompany the text:
 NUMERICAL METHODS for Mathematics, Science and Engineering, 2nd Ed, 1992
 Prentice Hall, Englewood Cliffs, New Jersey, 07632, U.S.A.
 Prentice Hall, Inc.; USA, Canada, Mexico ISBN 0-13-624990-6
 Prentice Hall, International Editions:   ISBN 0-13-625047-5
 This free software is compliments of the author.
 E-mail address:       in%"mathews@fullerton.edu"
 
 Algorithm 10.4 (Dirichlet Method for Laplace's Equation).
 Section   10.3, Elliptic Equations, Page 531
---------------------------------------------------------------------------
*/
/*
---------------------------------------------------------------------------
 
 Algorithm 10.4 (Dirichlet Method for Laplace's Equation)

                                               2
 To approximate the solution of u_(xx)(x,y) + u_(yy) (x,y)  =  0

 over R = {(x,y): 0 <= x <= a,  0 <= y <= b} with 

 u(x,0) = f_1(x),  u(x,b) = f_2(x)  for  0 <= x <= a  and 

 u(0,y) = f_3(y),  u(a,y) = f_4(y)  for  0 <= y <= b.

 It is assumed that  Delta(x) = Delta(y) = h  and that integers 

 n  and  m  exist so that  a = nh  and  b = mh.

 User has to supply functions  F1, F2, F3, F4  with gridpoints. 
 as arguments. An example is given in this program.

---------------------------------------------------------------------------
*/

#include<stdio.h>
#include<stdlib.h>
#include<math.h>

#define Max 50

/* Global Variables */

double H;                /* Step size  */

/* Prototypes */

double F1i(int i);
double F2i(int i);
double F3i(int i);
double F4i(int i);

/* Grid function for amplitude */

double F1i(int i)
{
    extern double H;
    double arg;

    arg = H * ( i - 1.0 );

    if( (arg >= 0) && (arg <= 4.0) )  return ( 180.0 );
    else
    {
        printf(" F1i() :Argument not in range ! Exiting !\n");
        printf(" arg  : %lf\n", arg);
        exit(0);
    }
}

double F2i(int i)
{
    extern double H;
    double arg;

    arg = H * ( i - 1.0 );

    if( (arg >= 0) && (arg <= 4) ) return ( 20.0 );
    else
    {
        printf(" F2i() :Argument not in range ! Exiting !\n");
        printf(" arg  : %lf\n", arg);
        exit(0);
    }
}

double F3i(int i)
{
    extern double H;
    double arg;

    arg = H * ( i - 1.0 );

    if( (arg >= 0) && (arg <= 4) ) return ( 80.0 );
    else
    {
        printf(" F3i() :Argument not in range ! Exiting !\n");
        printf(" arg  : %lf\n", arg);
        exit(0);
    }
}

double F4i(int i)
{
    extern double H;
    double arg;

    arg = H * ( i - 1.0 );

    if( (arg >= 0) && (arg <= 4) ) return ( 0.0 );
    else
    {
        printf(" F4i() :Argument not in range ! Exiting !\n");
        printf(" arg  : %lf\n", arg);
        exit(0);
    }
}


/* -------------------------------------------------------- */

/*  Main program for algorithm 10-4  */


/*  In  C  the fields begin with index 0. 
    In this program it is pretended that they begin with 1.
    So be careful with U[i][j] with i or j = 0.  */

void main(void)

{
    int i, j;                /* Loop Counters                         */
    double A, B;             /* INPUT : Width and height of Rectangle */
    double Ave;              /* INPUT : Initial approximation         */
    int N, M;                /* dimensions of the grid                */
    double U[Max][Max];      /* Grid-Amplitudes                       */
    int Count;               /* counter for while loop                */
    double Tol;              /* stop criteria                         */
    double w, Pi;
    double H;                /* Grid spacing                          */
    double Relax, temp;

    printf("-------------------------------------------------------\n");
    printf("------- Dirichlet Method for Laplace's Equation -------\n");
    printf("--------- Try Example 10.6 on page 528/529-------------\n");
    printf("-------------------------------------------------------\n");

    printf("Please enter A (interval boundary of x)\n");
    printf("For present example type  4\n");
    scanf("%lf",&A);
    printf("Please enter B (interval boundary of y)\n");
    printf("For present example type  4\n");
    scanf("%lf",&B);
    printf("You entered A = %lf and B = %lf \n", A, B);
    printf("------------------------------------------------\n");
    printf("Please enter Grid-Spacing H\n");
    printf("For present example type  0.5\n");
    scanf("%lf",&H);
    printf("You entered H = %lf\n", H);
    printf("------------------------------------------------\n");
    printf("\n");
    printf("Please enter dimension of grid in x-direction :\n");
    printf("For present example type  9\n");
    scanf("%d",&N);
    printf("Please enter dimension of grid in y-direction :\n");
    printf("For present example type  9\n");
    scanf("%d",&M);
    printf("You entered N = %d and M = %d\n", N, M);
    printf("------------------------------------------------\n");
    printf("\n");
    printf("Please enter the initial approximation :\n");
    printf("For present example type  70\n");
    scanf("%lf",&Ave);
    printf("You entered = %lf\n", Ave);
    printf("------------------------------------------------\n");
    printf("\n");

    /* Compute step sizes */

    Pi = 3.1415926535;

    /* Only to make the 0-indexes save we initialize them to zero.
       This is only because in C fields begin with index 0. 
       In this program we ignore the index 0. */

    for ( i = 0; i < N; i++ ) U[i][0] = 0.0;
    for ( i = 1; i < M; i++ ) U[0][i] = 0.0;

    /* Initialize starting values at the interior points */

    for ( i = 2; i <= N-1; i++ )
    {
        for ( j = 2; j <= M-1; j++ )   U[i][j] = Ave;
    }

    /* Store boundary values in the solution matrix */

    for ( j = 1; j <= M ; j++ ) 
    {
        U[1][j] = F3i(j);
        U[N][j] = F4i(j);
    }

    for ( i = 1; i <= N ; i++ ) 
    {
        U[i][1] = F1i(i);
        U[i][M] = F2i(i);
    }

    /* The SQR parameter */

    temp =  cos( Pi/(N-1) ) + cos( Pi/(M-1) );

    w = 4.0 / ( 2.0 + sqrt( 4.0 - temp * temp ) );

    /* Initialize the loop control parameters */

    Tol   = 1.0;
    Count = 0.0;

    while ( (Tol > 0.001) && (Count <= 140) )
    {
       Tol = 0.0;
       for ( j = 2; j <= M - 1; j++ )
       {
           for ( i = 2; i <= N - 1; i++ )
           {
             Relax = w * ( U[i][j+1] + U[i][j-1] + U[i+1][j] +
                           U[i-1][j] - 4.0 * U[i][j] ) / 4.0;
             U[i][j] += Relax;
             if( fabs(Relax) > Tol ) Tol = fabs(Relax);
           }
           Count++;
       }
    }

    /* Output the solution */

    for ( j = 1; j <= M; j++ )
    {
        for ( i = 1; i <= N; i++ ) printf("%8.4lf ", U[i][j]);
        printf("\n");
    }

}   /* End of main program */

