/*
---------------------------------------------------------------------------
 Algo7-5.c   C program for implementing Algorithm 7.5
 Algorithm translated to  C  by: Dr. Norman Fahrer
 IBM and Macintosh verification by: Daniel Mathews
 
 NUMERICAL METHODS: C Programs, (c) John H. Mathews 1995
 To accompany the text:
 NUMERICAL METHODS for Mathematics, Science and Engineering, 2nd Ed, 1992
 Prentice Hall, Englewood Cliffs, New Jersey, 07632, U.S.A.
 Prentice Hall, Inc.; USA, Canada, Mexico ISBN 0-13-624990-6
 Prentice Hall, International Editions:   ISBN 0-13-625047-5
 This free software is compliments of the author.
 E-mail address:       in%"mathews@fullerton.edu"
 
 Algorithm 7.5 (Adaptive Quadrature Using Simpson's Rule).
 Section   7.4, Adaptive Quadrature, Page 389
 This program is to be used for pedagogical purposes.
---------------------------------------------------------------------------
*/
/*
---------------------------------------------------------------------------

 Algorithm 7.5 (Adaptive Quadrature using Simpson's Rule).

 To  approximate the integral

     b
     /               M   h_k
     | f(x)dx   =   Sum  --- [ f(x_(4k-3)) + 2*f(x_(4k-2))
     /              k=1   3
     a                         + f(x_(4k-1)) + f(x_(4k)) ].

 The composite Simpson rule is applied to the  4M  subintervals
 [x_(4k-4),x_(4k)],  where  [A,B] = [x_0, x_(4M)]  and
  x_(4k-4+j) = x_(4k-4) + jh_k 
 for each  k = 1,2,...,M  and  j = 1,2,3,4.

---------------------------------------------------------------------------
*/

/* User has to supply a function named : ffunction
   An example is included in this program. See page
   387/388 example 7.16 in the book.  */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define NMAX 200
#define MMAX 12

/* Prototypes */

int SimpsonRule(double GetA, double GetB, double GetTol);
int REFINE(int P);

/* Global variables */

double Err,Tol,Check;
int State, M, N;
int Iterating = 0;
int Done      = 1;
double SR[NMAX][MMAX];
double SRTemp[NMAX];
double Tol = 1E-6; 

/*  Meaning of the variables : 

    SRTemp[7], SRTemp[8]         Simpson's rule approximations
    SRTemp[4,5,6]                Function values
    Err,Tol,Check                Variables to do checks
    SR[][]                       Simpson Rule result records
    double Tol                   Interval tolerance

*/

/*  define prototype for USER-SUPPLIED function f(x)  */

    double ffunction(double x);

/*  EXAMPLE for "ffunction" 
    See page 387/388 example No. 7.16 in the book  */

    double ffunction(double x)

    {
        double fact;

        fact = -1.5 * x;
        return ( 13.0 * ( x - x * x) * exp(fact) );
    }

/* -------------------------------------------------------- */

/*  Main program for algorithm 7.5  */

void main()
{
    /* local variables */

    int J;                     /* loop counters                  */
    double A, B;               /* Endpoints of the interval      */
    double C;                  /* A + half of interval width     */

    double SUM;                /* evolving integral value        */
    double SumErr;             /* Error bound                    */          

    printf("----------------------------------------------\n");
    printf("----Adaptive Quadrature Using Simpson's Rule--\n");
    printf("----Example 7/16 from page 387/388------------\n");
    printf("----------------------------------------------\n");
    printf("----------------------------------------------\n");
    printf("Please enter the lower boundary of the integral: \n");
    printf("In this Example : A = 0 , so type: 0 \n");
    scanf("%lf",&A);   
    printf("Please enter the lower boundary of the integral: \n");
    printf("In this Example : B = 4 , so type: 4\n");
    scanf("%lf",&B);
    printf("The boundaries of the integral are: %lf %lf\n",A, B);
    printf("----------------------------------------------\n");

    SimpsonRule(A,B,Tol);

    /* Fill Row 1 of matrix SR */    

    SR[1][1]  = SRTemp[1];
    SR[1][2]  = SRTemp[2];
    SR[1][3]  = SRTemp[3];
    SR[1][4]  = SRTemp[4];
    SR[1][5]  = SRTemp[5];
    SR[1][6]  = SRTemp[6];
    SR[1][7]  = SRTemp[7];
    SR[1][8]  = SRTemp[8];
    SR[1][9]  = SRTemp[9];
    SR[1][10] = SRTemp[10];
    SR[1][11] = SRTemp[11];

    M = 1;

    State = 0;

    while (State == Iterating)
    {
      N = M;

      for( J = N; J >= 1; J-- )  REFINE(J);
    }

    SUM = SumErr = 0;

    for ( J = 1; J <= M; J++ ) 
    {
        SUM    +=      SR[J][8] ;
        SumErr += fabs(SR[J][9]) ;
    }

    printf("----------------------------------------------\n");

    printf(" The approximate  value of the integral\n");
    printf(" is : %16.10lf\n", SUM );
    printf("----------------------------------------------\n");
    printf(" The Error bound is :\n");
    printf(" is : %12.10lf\n", SumErr );
    printf("----------------------------------------------\n");


}   /* End of main programm */

/* Subroutine for simpson rule */

int SimpsonRule(double GetA, double GetB, double GetTol)
{
    double H;
    int i;

    SRTemp[1]  =  GetA;
    SRTemp[3]  =  GetB;
        H      =  (SRTemp[3] - SRTemp[1]) / 2.0;
    SRTemp[2]  =  SRTemp[1] + H;
    SRTemp[4]  =  ffunction(SRTemp[1]);
    SRTemp[5]  =  ffunction(SRTemp[2]);
    SRTemp[6]  =  ffunction(SRTemp[3]);
    SRTemp[7]  =  H * (SRTemp[4] + 4.0 * SRTemp[5] + SRTemp[6]) / 3.0 ;
    SRTemp[8]  =  SRTemp[7];
    SRTemp[9]  =  Err    = GetTol;
    SRTemp[10] =  Tol    = GetTol;
    SRTemp[11] =  Check  = 0.0;

    return;
}

/* Subroutine to refine the SR[][1,..,11] vectors */

int REFINE(int P)
{
    double SR0[MMAX];
    double SR1[MMAX];
    double SR2[MMAX];
    double Tol2;
    double A, B, C;
    int i, J;

    State = Done;

    /* Vector replacements */

    for (i = 1; i <= MMAX-1; i++) SR0[i] = SR[P][i];

    A       =  SR0[1]  ;
    C       =  SR0[2]  ;
    B       =  SR0[3]  ;
    Err     =  SR0[9]  ;
    Tol     =  SR0[10] ;
    Check   =  SR0[11] ;

    /* Return if refinement is not needed */

    if ( (int)(Check) == 1 ) return ;

    /* Bisect the interval and apply Simpson's rule 
       recursively */

    Tol2 = Tol / 2.0;
    SimpsonRule(A,C,Tol2);


    for (i = 1; i <= MMAX-1; i++) SR1[i] = SRTemp[i];

    SimpsonRule(C,B,Tol2);

    for (i = 1; i <= MMAX-1; i++) SR2[i] = SRTemp[i];

    /* Determine the error bound and 
       if the interval meets tolerance */
 
    Err = ( fabs(SR0[7] - SR1[7] - SR2[7]) ) / 10.0;
    
    
    if ( Err < Tol ) 
    {

     /* Set 'Check' to 1 so next entry does RETURN */

     SR0[11] = 1.0;

     /* Vector replacements */

     for (i = 1; i <= MMAX-1; i++) SR[P][i] = SR0[i];

     /* Record Simpson composite rule and the error bound */
     
     SR[P][8]  = SR1[7] + SR2[7];
     SR[P][9]  = Err;
    }

    else 
    {

      /* Shift the list of records for the subintervals,
         and quadrature information and insert the new
         vector records for the bisected interval */
      
      for(J = M+1; J >= P; J--)
      {
         for (i = 1; i <= MMAX-1; i++) SR[J][i] = SR[J-1][i];
      }
      
      M++;
      
      for (i = 1; i <= MMAX-1; i++) SR[P][i]   = SR1[i];
      for (i = 1; i <= MMAX-1; i++) SR[P+1][i] = SR2[i];
     
      State = Iterating;
      
    }
  
    return;
}

