/*
---------------------------------------------------------------------------
 Alog8-3.c   C program for implementing Algorithm 8.3
 Algorithm translated to  C  by: Dr. Norman Fahrer
 IBM and Macintosh verification by: Daniel Mathews
 
 NUMERICAL METHODS: C Programs, (c) John H. Mathews 1995
 To accompany the text:
 NUMERICAL METHODS for Mathematics, Science and Engineering, 2nd Ed, 1992
 Prentice Hall, Englewood Cliffs, New Jersey, 07632, U.S.A.
 Prentice Hall, Inc.; USA, Canada, Mexico ISBN 0-13-624990-6
 Prentice Hall, International Editions:   ISBN 0-13-625047-5
 This free software is compliments of the author.
 E-mail address:       in%"mathews@fullerton.edu"
 
 Algorithm 8.3 (Local Minimum Search: Quadratic Interpolation).
 Section   8.1, Minimization of a Function, Page 416
---------------------------------------------------------------------------
*/
/*
---------------------------------------------------------------------------

 Algorithm 8.3 (Local Minimum Search Using Quadratic Interpolation).

 To find a local minimum of the function  f(x)  over the interval
 [a,b], by starting with one initial approximation  p_0  and then
 searching the intervals  [a,p_0]  and  [p_0,b].

---------------------------------------------------------------------------
*/

/* User has to supply a function named : ffunction and its
   derivative dffunction.

   An example is included in this program */

#include<stdio.h>
#include<stdlib.h>
#include<math.h>


/*  define prototype for USER-SUPPLIED function f(x)  */

    double ffunction(double x);
    double dffunction(double x);


/*  EXAMPLE for "ffunction" : see page 420 No. 10  */

    double ffunction(double x)

    {
        return ( -sin(x) - sin(3 * x) / 3.0 );
    }

/*  EXAMPLE for "dffunction"  */

    double dffunction(double x)

    {
        return ( - cos(x) - cos(3 * x) );
    }

/* -------------------------------------------------------- */

/*  Main program for algorithm 8.3  */


void main(void)
{
    double Delta   = 1E-12;   /* Convergence tolerance for abscissas */
    double Epsilon = 1E-14;   /* Convergence tolerance for ordinates */
    int Jmax = 20;            /* Maximum number of iterations        */
    int Kmax = 50;            /* Maximum number of iterations        */
    int K, J, Cond;
    double P0, P1, P2;
    double Y0, Y1, Y2;
    double D, Err, H;
    double Hmin, Pmin, Ymin;

    printf("------------------------------------------------\n");
    printf("====== Example/Problem No. 10 on page 420 ======\n");
    printf("------------------------------------------------\n");
    printf("Please enter starting value : P0\n");
         printf("For present example type : 1.0\n");
    scanf("%lf", &P0);
         printf("You entered P0 = %lf\n", P0);    
    printf("\n");

    /* Initialize variable, step size and counter */

    Err = 1.0;
    H = 1.0;
    K = 0;

    /* Loop to find Ymin */
        
    while ( (K < Kmax) && (Err > Epsilon) )
    {

      if( dffunction(P0) > 0 ) H = - fabs(H); /* Slope points downhill */
      P1 = P0 + H;             /* Two more test points        */
      P2 = P0 + 2.0 * H;       
      Y0 = ffunction(P0);      /* Compute the function values */
      Y1 = ffunction(P1);
      Y2 = ffunction(P2);
      Cond = 0;                /* Initialize the variables    */
      J    = 0;
      
      /* Determine H so that Y1 < Y0 and Y1 < Y2 */

      while( (J < Jmax) && (H > Delta) && (Cond == 0 ) ) 
      {

         if ( Y0 <= Y1 )
         {
            P2 = P1;
            Y2 = Y1;
            H  = H / 2.0;            /* Make H smaller */
            P1 = P0 + H;
            Y1 = ffunction(P1);
         } 

         else if (Y2 < Y1)
         {
               P1 = P2;
               Y1 = Y2;   
               H  = 2.0 * H;
               P2 = P0 + 2.0 *H;     /* Make H larger  */
               Y2 = ffunction(P2);
         }

         else Cond = -1;             /* Loop termination */
         

      J++;                           /* Increment the counter */

      }  /* end of minor while-loop ( proper H is now determined) */

      /* Use quadratic interpolation to compute the abscissa at the
         minimum. Also check division by zero. */      
   
      D = 4.0 * Y1 - 2.0 * Y0 - 2.0 * Y2;
      
      if( D < 0 )
      {
         Hmin = H * ( 4.0 * Y1 - 3.0 * Y0 - Y2 ) / D;
      }
      else
      {
         Hmin = H / 3.0;
         Cond = 4;
      }

      Pmin = P0 + Hmin;               /* Coordinates of the minimum */
      Ymin = ffunction(Pmin);
      H = fabs(H);

      /* How determine the magnitude of the next H */

      if( fabs(Hmin) < H )  H = fabs(Hmin);

      if ( fabs(Hmin - H) < H )  H = fabs(Hmin - H);

      if ( fabs(Hmin - 2.0 * H) < H )  H = fabs(Hmin - 2.0 * H);

      if( H  < Delta )  Cond = 1;     /* Convergence of abscissa */      

      Err = fabs(Y0 - Ymin);
       
      if( fabs(Y1 - Ymin) < Err ) Err = fabs(Y1 - Ymin);
      else if( fabs(Y2 - Ymin) < Err ) Err = fabs(Y2 - Ymin);

      if( Err < Epsilon ) Cond = 2;   /* Convergence of ordinate */

      K++;
      P0 = Pmin;                      /* Update the value P0 */
      
      if ( (Cond = 2) && ( H < Delta) ) Cond = 3;

   }     /* end of major while-loop ( Ymin is found by now. )*/

    /* OUTPUT */

    printf("--------------------------------------------------\n");
    printf("An approximation for the minimum is : %lf\n", Pmin);
    printf("An estimate for its accuracy is : %lf\n", fabs(H));
    printf("The function value f(Pmin) is : %lf\n", Ymin);
    printf("--------------------------------------------------\n");

    if ( Cond == 0 ) 
    {
       printf("Convergence is doubtful because the maximum number of\n");
       printf("iterations was exceeded\n");
    }

 if ( Cond == 1 ) printf("Convergence of the abscissas has been achieved\n");
 if ( Cond == 2 ) printf("Convergence of the ordinates has been achieved\n");
 if ( Cond == 3 ) printf("Convergence of both coordinates has been achieved\n");

    if ( Cond == 4 ) 
    {
       printf("Convergence is doubtful because division by zero\n");
       printf("was encountered\n");
    }
    

} 

