/*
---------------------------------------------------------------------------
 Algo8-4.c   C program for implementing Algorithm 8.4
 Algorithm translated to  C  by: Dr. Norman Fahrer
 IBM and Macintosh verification by: Daniel Mathews
 
 NUMERICAL METHODS: C Programs, (c) John H. Mathews 1995
 To accompany the text:
 NUMERICAL METHODS for Mathematics, Science and Engineering, 2nd Ed, 1992
 Prentice Hall, Englewood Cliffs, New Jersey, 07632, U.S.A.
 This free software is compliments of the author.
 E-mail address:       in%"mathews@fullerton.edu"
 
 Algorithm 8.4, (Steepest Descent or Gradient Method).
 Section   8.1, Minimization of a Function, Page 418
---------------------------------------------------------------------------
*/
/*
---------------------------------------------------------------------------

 Algorithm 8.4 (Steepest Descent or Gradient Method).

                                                  ->
 To numerically approximate a local minimum of  F(X ), where  F
 is a continuous function of  N  real variables 
 ->  
 X  =  (x_1, x_2, ..., x_N), by starting with one point
 
 ->
 P_0 and using the gradient method.

---------------------------------------------------------------------------
*/

/* User has to supply a function named : ffunction and its
   gradient along with the unit vector pointing towards
   the minimum.

   An example is included in this program */

#include<stdio.h>
#include<stdlib.h>
#include<math.h>

#define MAX   10
#define Jmax  20   /* Maximum number of iterations */
#define Max   50   /* Maximum number of iterations */


/*  global variables */

    double P0[MAX], Pmin[MAX];
    double Ymin, Y0;
    double H   = 1.0;
    double Err = 1.0;            
    double S[MAX], G[MAX];
    int Cond;


/*  define prototype for USER-SUPPLIED function f(x)  */

    double ffunction(double *P);
    void Gradient(double *P);
    void  Quadmin(int NC, double Delta, double Epsilon);


/*  EXAMPLE for "ffunction" : see page 422 No. 34  */

    double  ffunction(double *P)
    {
        return ( pow(P[0],4) + pow(P[1],4) + pow(P[2],4) - 
                 4.0 * P[0] * P[1] * P[2] );
    }

/*  EXAMPLE for "Gradient"  */

    void Gradient(double *P)
    {

        /* gradient */

        G[0] = 4.0 * ( pow(P[0],3) - P[1]*P[2] );
        G[1] = 4.0 * ( pow(P[1],3) - P[0]*P[2] );
        G[2] = 4.0 * ( pow(P[2],3) - P[0]*P[1] );

        /* unit vector along gradient */
        
        S[0] = - G[0] / sqrt( pow(G[0],2) + pow(G[1],2) + pow(G[2],2) );
        S[1] = - G[1] / sqrt( pow(G[0],2) + pow(G[1],2) + pow(G[2],2) );
        S[2] = - G[2] / sqrt( pow(G[0],2) + pow(G[1],2) + pow(G[2],2) );
    }
    

/* -------------------------------------------------------- */

/*  Main program for algorithm 8.4  */

void main(void)
{
    double Delta   = 1E-5;    /* Convergence tolerance for points    */
    double Epsilon = 1E-7;    /* Tolerance for the function values   */
    int NC;                   /* Number of components of gradient    */
    int Count = 0, i;

    printf("------------------------------------------------\n");
    printf("------Steepest Descent or Gradient Method-------\n");
    printf("------------------------------------------------\n");
    printf("====== Example/Problem No. 34 on page 422 ======\n");
    printf("------------------------------------------------\n");
    printf("\n");
    printf("Give the number of coordinates for the starting point:\n");
    printf("For this example we have  3  coordinates.\n");
    printf("( Not more than 10 )\n");
    scanf("%d", &NC); 
    printf("You say there are %d components.\n");

    if( NC > MAX ) 
    { 
     printf("Not more than %d components for starting point allowed.\n",MAX);
     printf("You can change the program to accept more. For now : terminating..\n");
    }


    printf("------------------------------------------------\n");
    for (i = 1; i <= NC; i++)
    {
     printf("Please enter %d st/nd/rd/th coordinate of starting point :\n", i);
     scanf("%lf", &P0[i-1]);
     printf("You entered  %lf\n", P0[i-1]);    
     printf("\n");
    }

    Y0 = ffunction(P0);

    while ( (Count < Max) && (H < Delta) || (Err > Epsilon) )
    {
      Gradient(P0);
      Quadmin(NC, Delta, Epsilon);

      for ( i = 0; i < NC; i++ )  P0[i] = Pmin[i];
      
      Y0 = Ymin;
      
      Count++;
    }     

    /* OUTPUT */

    printf("--------------------------------------------------\n");
    printf("The local minimum is at:\n");
    for ( i = 0; i < NC; i++ ) printf("%12.7lf ", Pmin[i]);
    printf("\n");

    printf("The minimum value of the function is : %lf\n", Ymin);
    printf("--------------------------------------------------\n");

    if ( Cond == 0 ) 
    {
       printf("Convergence is doubtful because the maximum number of\n");
       printf("iterations was exceeded\n");
    }

    if ( Cond == 1 ) printf("Convergence of the abscissas has been achieved\n");

    if ( Cond == 2 ) printf("Convergence of the ordinates has been achieved\n");

    if ( Cond == 3 ) printf("Convergence of both coordinates has been achieved\n");

    if ( Cond == 4 ) 
    {
       printf("Convergence is doubtful because division by zero\n");
       printf("was encountered\n");
    }


} /* End of Main */


/* Procedure Quadmin */

void  Quadmin(int NC, double Delta, double Epsilon)
{
   double P1[MAX];
   double P2[MAX];
   double H0, H1, H2, Hmin, E0, E1, E2, Y1, Y2, D;
   int i, J;

      
   Cond = 0;
   J    = 0;

   for ( i = 0; i < NC; i++ )
   {
      P1[i] = P0[i] + H * S[i]; 
      P2[i] = P0[i] + 2.0 * H * S[i]; 
   }

   Y1 = ffunction(P1);
   Y2 = ffunction(P2);

   /* Determine H so that Y1 < Y0 and Y1 < Y2 */

   while ( (J < Jmax) && (Cond = 0) )
   {
      if ( Y0 <= Y1 )   /* Make H smaller */
      {
         Y2    = Y1;
         H     = H / 2.0;
         for ( i = 0; i < NC; i++ )
         {
            P2[i] = P1[i];
            P1[i] = P0[i] + H * S[i];
         }   
         Y1 = ffunction(P1);
      }

      else
      { 
         if ( Y2 < Y1 )  /* Make H larger */
         {
            Y1 = Y2;
            H  = 2.0 * H;         
            for ( i = 0; i < NC; i++ )
            {
               P1[i] = P2[i];
               P2[i] = P0[i] + 2.0 * H * S[i];
            }
            Y2 = ffunction(P2);
         }
         else Cond = -1;
      }
   } /* end of While-loop */


   if ( H < Delta ) Cond = 1;
   D = 4.0 * Y1 - 2.0 * Y0 - 2.0 * Y2;

   /* Quadratic interpolation to find Hmin */

   if ( D < 0 ) Hmin = H * ( 4.0 * Y1 - 3.0 * Y0 - Y2 ) / D;
   else    /* check division by zero */
   {
      Cond = 4;
      Hmin = H / 3.0;
   }
   
   for ( i = 0; i < NC; i++ ) Pmin[i] = P0[i] + Hmin * S[i];
   
   Ymin = ffunction(Pmin);

   /* Convergence test for the points */

   H0 = fabs(Hmin);
   H1 = fabs(Hmin - H);
   H2 = fabs(Hmin - 2.0 * H);

   if ( H0 < H )    H    = H0;
   if ( H1 < H )    H    = H1;
   if ( H2 < H )    H    = H2;
   if ( H < Delta ) Cond = 1;

   /* Convergence test for the function values */

   E0 = fabs(Y0 - Ymin);
   E1 = fabs(Y1 - Ymin);
   E2 = fabs(Y2 - Ymin);

   if      ( E0 < Err ) Err = E0;
   else if ( E1 < Err ) Err = E1;
   else if ( E2 < Err ) Err = E2;
   else if ( (E0 == 0) && (E1 == 0) && (E2 == 0) ) Err = 0;

   if ( Err < Epsilon )              Cond = 2;
   if ( (Cond == 2) && (H < Delta) ) Cond = 3;
   J++;

} /* end of procedure QUADMIN */   

