/*
---------------------------------------------------------------------------
 Algo9-5.c   C program for implementing Algorithm 9.5
 Algorithm translated to  C  by: Dr. Norman Fahrer
 IBM and Macintosh verification by: Daniel Mathews
 
 NUMERICAL METHODS: C Programs, (c) John H. Mathews 1995
 To accompany the text:
 NUMERICAL METHODS for Mathematics, Science and Engineering, 2nd Ed, 1992
 Prentice Hall, Englewood Cliffs, New Jersey, 07632, U.S.A.
 Prentice Hall, Inc.; USA, Canada, Mexico ISBN 0-13-624990-6
 Prentice Hall, International Editions:   ISBN 0-13-625047-5
 This free software is compliments of the author.
 E-mail address:       in%"mathews@fullerton.edu"
 
 Algorithm 9.5 (Runge-Kutta-Fehlberg Method RKF45).
 Section   9.5, Runge-Kutta Methods, Page 461
---------------------------------------------------------------------------
*/
/*
---------------------------------------------------------------------------
 
 Algorithm 9.5 (Runge-Kutta-Fehlberg Method (RKF45)).

 To approximate the solution of the initial value problem  y' = f(t,y)
 with  y(a) = y_0  over [a,b] with an error control and variable
 step-size method.

 User has to supply  functions named : ffunction 
 An example is included in this program.

---------------------------------------------------------------------------
*/

#include<stdio.h>
#include<stdlib.h>
#include<math.h>

#define MAX 500


/*  define prototype for USER-SUPPLIED function f(x)  */

    double ffunction(double t, double y);

/*  EXAMPLE for "ffunction"   
    In this example the function is actually independent of 
    the variable  t   */

    double ffunction(double t, double y)

    {
        return ( 1 + y * y );
    }

/* -------------------------------------------------------- */

/*  Main program for algorithm 9.5  */

void main(void)

{
    double Tol = 2 * 1E-09;    /* Error control tolerance        */
    int I, J;                  /* Loop counter                   */
    int N;                     /* Number of tentative steps      */
    double A, B, Y[MAX];       /* Endpoints and inital value     */
    double H;                  /* Compute the step size          */
    double T[MAX]; 
    double K1, K2, K3, K4, K5, K6;   /* Function values          */
    double t, y;
    double Hmin, Hmax;         /* Minimum and maximum step size  */
    double Err, S;             /* Error and step size scalar     */


    printf("----------- Taylor's Method -------------------\n");
    printf("--------- Example 9.12 on page 459 -------------\n");
    printf("-----------------------------------------------\n");
    printf("Please enter endpoints of the interval [A,B]:\n");
    printf("For used Example type  : 0, 1.4\n");
    scanf("%lf, %lf", &A, &B);
    printf("You entered [%lf, %lf]\n", A, B);
    printf("-----------------------------------------\n");
    printf("Please enter number of tentative steps: (Not more than %d !)\n", MAX
);
    scanf("%d",&N);
    if(N > MAX)
    {
    printf(" Not prepared for more than %d steps. Terminating. Sorry\n",MAX);
    exit(0);
    }
    printf("-----------------------------------------\n");


    printf("Please enter initial value Y[0] :\n");
    printf("For used Example type : 0\n");
    scanf("%lf", &Y[0]);
    printf("You entered Y[0] = %lf\n", Y[0]);

    /* Compute the step size */

    H    = (B - A) / N;
    Hmin = H / 64.0;
    Hmax = H * 64.0;      

    /* Initialize the variable */

    T[0] = A;
    J    = 0;

    while(T[J] < B)
    {
      if( (T[J] + H) > B ) H = B - T[J];   /* The last step */
      t  = T[J];
      y  = Y[J];

      /* Compute the function values */

      K1 = H * ffunction(t,y);

      K2 = H * ffunction(t + 0.25 * H, y + 0.25 * K1);

      K3 = H * ffunction(t + 3.0*H/8.0, y + 3.0*K1/32.0 + 9.0*K2/32.0);

      K4 = H * ffunction(t + 12.0*H/13.0, y + 1932.0*K1/2197.0 -
                        7200.0*K2/2197.0 + 7296.0*K3/2197.0);

      K5 = H * ffunction(t + H, y + 439.0*K1/216.0 - 8.0*K2
                         + 3680.0*K3/513.0  - 845.0*K4/4104.0);

      K6 = H * ffunction(t + 0.5*H, y - 8.0*K1/27.0 + 2.0*K2 -
                        3544.0*K3/2565.0 + 1859.0*K4/4104.0 - 11.0*K5/40.0);

      /* { Z_(K+1) - Y_(K+1) } */      
      
      Err = fabs( K1/360.0 - 128.0*K3/4275.0 - 2197.0*K4/75240.0
                 + K5/50.0 + 2.0*K6/55.0 );


      /* Accept the approximation */

      if ( (Err < Tol) || (H <= 2.0*Hmin) )
      {
         Y[J+1] = y + 25.0*K1/216.0 + 1408.0*K3/2565.0 + 
                  2197.0*K4/4104.0 - K5/5.0;

         T[J+1] = t + H;
         J++;
      }

      /* Trap division by zero */

      if( Err == 0 ) S = 0;

      /* Step size scalar */

      else S = 0.84 * sqrt( sqrt(Tol * H / Err) );

      /* Reduce or increase step */

      if ( (S < 0.75) && (H > 2.0*Hmin) )   H /=  2.0;
      else if ( (S > 1.50) && (2.0 * H < Hmax) ) H *=  2.0;
   
   } /* End of while loop */

    /* Output */

    for ( I = 0; I <= J; I++)
    {
       printf("I = %d, T[I] = %lf, Y[I] = %lf\n", I, T[I], Y[I]);
    } 

}   /* End of main program */

