/*
---------------------------------------------------------------------------
 Algo9-9.c   C program for implementing Algorithm 9.9
 Algorithm translated to  C  by: Dr. Norman Fahrer
 IBM and Macintosh verification by: Daniel Mathews
 
 NUMERICAL METHODS: C Programs, (c) John H. Mathews 1995
 To accompany the text:
 NUMERICAL METHODS for Mathematics, Science and Engineering, 2nd Ed, 1992
 Prentice Hall, Englewood Cliffs, New Jersey, 07632, U.S.A.
 Prentice Hall, Inc.; USA, Canada, Mexico ISBN 0-13-624990-6
 Prentice Hall, International Editions:   ISBN 0-13-625047-5
 This free software is compliments of the author.
 E-mail address:       in%"mathews@fullerton.edu"
 
 Algorithm 9.9 (Linear Shooting Method).
 Section   9.8, Boundary Value Problems, Page 488
---------------------------------------------------------------------------
*/
/*
---------------------------------------------------------------------------
 
 Algorithm 9.9 (Linear Shooting Method)

 To approximate the solution of the boundary value problem 

 x''  =  p(t)*x'(t) + q(t)*x(t) + r(t)  with  x(a) = Alpha

 and  x(b) = Beta  over the interval [a,b] by using the Runge-
 Kutta method of order 4.

 User has to supply  functions named : ffunction 
 An example is included in this program.

---------------------------------------------------------------------------
*/

#include<stdio.h>
#include<stdlib.h>
#include<math.h>

#define MAX 300


/*  define prototype for USER-SUPPLIED function f(x)  */

    double ffunction(int F1orF2, double t, double x, double y);

/*  EXAMPLE for "ffunction"   */

    double ffunction(int F1orF2, double t, double x, double y)

    {
        if( F1orF2 == 0 )
        {
            return ( 2.0*t*y/(1.0 + t*t) -  2.0*x/(1.0 + t*t) + 1.0 );
        }
        else
        {
            return ( 2.0*t*y/(1.0 + t*t) -  2.0*x/(1.0 + t*t) );
        }
    }

/* -------------------------------------------------------- */

/*  Main program for algorithm 9.9  */

void main(void)

{
    int K, J;                  /* Loop counter                   */
    int N;                     /* Number of Steps                */
    double A, B;               /* Endpoints and inital value     */
    double H;                  /* Compute the step size          */
    double T[MAX]; 
    double K1, K2, K3, K4;     /*                                */
    double R1, R2, R3, R4;     /*                                */
    int F1orF2 = 0;            /* Switch for function choice     */
    double Xj, Yj, Tj;         /*                                */           
    double X[MAX], Y[MAX];
    double X1[MAX], X2[MAX];
    double Alpha, Beta;        /* Boundary values                */

    printf("-----------Linear Shooting Method---------------\n");
    printf("--------- Example 9.17 on page 486 -------------\n");
    printf("-----------------------------------------------\n");
    printf("Please enter endpoints of the interval [A,B]:\n");
    printf("For used Example type  : 0, 4.0\n");
    scanf("%lf, %lf", &A, &B);
    printf("You entered [%lf, %lf]\n", A, B);
    printf("-----------------------------------------\n");
    printf("Please enter number of steps: (  < %d !)\n", MAX+1);
    scanf("%d",&N);
    if( N > MAX )
    {
    printf(" Number of steps must be less than %d\n",MAX+1);
    printf(" Terminating. Sorry\n");
    exit(0);
    }
    printf("-----------------------------------------\n");

    printf("Please enter Boundary values Y[A], Y[B]:\n");
    printf("Example 9.17 page 486:  1.25, -0.95\n");
    scanf("%lf, %lf", &Alpha , &Beta);

    printf("You entered Y[A] = %lf\n", Alpha);
    printf("You entered Y[B] = %lf\n", Beta);

    H    = (B - A) / N;

    /* Do what subroutine RKbdd4 does (twice) */

    for ( K = 0; K <= 1; K++)
    {
        /* Runge-Kutta order 4 */

        F1orF2 = K;

        if( K == 0 )
        {
         X[0]  = Alpha;
         X1[0] = X[0];
         Y[0]  = 0.0;
        }
        else
        {
         X[0]  = 0.0;
         X2[0] = X[0];
         Y[0]  = 1.0;
        }

        T[0] = A;

        for( J = 0; J <= N - 1; J++)
        {
            Tj = T[J];
            Xj = X[J];
            Yj = Y[J];
            K1 = H * Yj;
            R1 = H * ffunction(F1orF2, Tj, Xj, Yj);
            K2 = H * ( Yj + R1 / 2.0);
            R2 = H * ffunction(F1orF2, Tj + H/2.0, Xj + K1/2.0, Yj + R1/2.0);
            K3 = H * ( Yj + R2 / 2.0);
            R3 = H * ffunction(F1orF2, Tj + H/2.0, Xj + K2/2.0, Yj + R2/2.0);
            K4 = H * (Yj + R3);
            R4 = H * ffunction(F1orF2, Tj + H , Xj + K3, Yj + R3);
            X[J+1] = Xj + (K1 + 2.0*K2 + 2.0*K3 + K4) / 6.0;
            Y[J+1] = Yj + (R1 + 2.0*R2 + 2.0*R3 + R4) / 6.0;
            T[J+1] = A + H * (J+1);

            if(K == 0) X1[J+1] = X[J+1]; 
            if(K == 1) X2[J+1] = X[J+1]; 
            
        }


    }      /* End of for loop to cycle twice through algorithm */



    for (J = 0; J <= N; J++) X[J] = X1[J] + (Beta - X1[N]) * X2[J] / X2[N];
    
    /* Output */

    for ( J = 0; J <= N; J++)
    {
       printf("J = %d, T[J] = %lf, X[J] = %lf\n", J, T[J], X[J]);
    } 

}   /* End of main program */

